""" Find artifacts in signals using spectral energy.

A period is considered an artifact when the energy of a single
frequency contains more than ("Threshold" * total energy in period).
The length of the considered periods can be set using the "FFT samples"
option. The "Cutoff frequency" determines the maximum frequency at which
artifacts are expected to occur. The "Overlap percentage" can be used to
achieve more fine grained artifact borders at the cost of performance.

Finally, a heuristic is employed to merge and/or drop artifact epochs:
First, artifact periods up to ("Maximum merging distance" * "FFT samples")
samples apart are merged into single epochs. Next, epochs shorter than
("Minimum artifact length" * "FFT samples") are not considered artifacts
(this is useful when the artifact energy fluctuates stronly).

Neo epochs are created and added to the correct segment. Each epoch is
labeled with the "Epoch label" and gets an annotation 'tag' defined 
by the "Annotation Tag".
"""
from spykeutils.plugin import analysis_plugin, gui_data

import scipy as sp
import quantities as pq
import neo

from botmpy.nodes import SpectrumArtifactDetector


class SpectralArtifactDetectionPlugin(analysis_plugin.AnalysisPlugin):
    nfft_index = (32, 64, 128, 256, 512, 1024, 2048, 4096, 8192)
    nfft_names = (str(s) for s in nfft_index)
    tag = gui_data.StringItem('Epoch Annotation tag')
    label = gui_data.StringItem('Epoch label', default='Artifact')
    fft_samples = gui_data.ChoiceItem('FFT samples', nfft_names, 
                                      default=4)
    threshold = gui_data.FloatItem('Threshold', default=0.25)
    cutoff = gui_data.FloatItem('Cutoff frequency', default=2000.0,
                                   unit='Hz')
    overlap = gui_data.ChoiceItem('Overlap percentage', ('0', '50', '75'),
                                  default=1)
    max_merge = gui_data.IntItem('Maximum merging distance', default=6)
    min_allowed = gui_data.IntItem('Minimum artifact length', default=2)
    
    def get_name(self):
        return 'Spectral Artifact Detection'

    def start(self, current, selections):
        current.progress.begin('Detecting artifacts...')
        all_signals = current.analog_signals_by_segment()
        
        detector = SpectrumArtifactDetector(
            nfft=self.nfft_index[self.fft_samples],
            threshold_factor=self.threshold,
            cutoff_hz=self.cutoff,
            en_func='max_normed',
            overlap=self.overlap,
            max_merge_dist=self.max_merge,
            min_allowed_length=self.min_allowed)
            
        current.progress.set_ticks(len(all_signals))
        for s, sigs in all_signals.iteritems():
            current.progress.set_status(s.name)
            
            detector.srate = sigs[0].sampling_rate.rescale(pq.Hz)
            l = sigs[0].shape[0]
            
            for sig in sigs:
                if not sig.sampling_rate == detector.srate:
                    raise SpykeException(
                        'All signals from a segment need to have the same'
                        'sampling rate!')
                if not sig.shape[0] == l:
                    raise SpykeException(
                        'All signals from a segment need to have the same'
                        'number of samples!')
            
            # Use SpectrumArtifactDetector to find artifacts
            detector.reset()
            detector(sp.vstack([sp.asarray(sig) for sig in sigs]).T)
            
            # Create and attach neo Epochs
            t = 1.0 / detector.srate
            for e in detector.events:
                ep = neo.Epoch(t * e[0], t * (e[1] - e[0]), 
                               self.label)
                if self.tag:
                    ep.annotate(tag=self.tag)
                s.epochs.append(ep)
                ep.segment = s
            current.progress.step()
                
        current.progress.done()